# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from jax._src import deprecations

# Added February 16, 2024.
_msg = ("Importing the jax.config submodule via `import jax.config` is deprecated."
        " To configure JAX use `import jax` and then reference the config object"
        " via `jax.config`.")
if deprecations.is_accelerated("jax.config", "config-module"):
  raise ImportError(_msg)
else:
  warnings.warn(_msg, DeprecationWarning, stacklevel=2)
del deprecations, warnings, _msg
